/* mask-seq.c -- mask out portions of a sequence
 *
 * This program reads a sequence file in FastA format and a file
 * containing position intervals, and produces a new version of the
 * sequence file with the specified intervals masked out, either by
 * changing the nucleotide characters to lowercase (default) or by
 * replacing them with a specified mask character (e.g., 'X' or 'N').
 * The position file will typically be a repeats file, but could
 * also be in the format for exons, annotations, underlays, or any
 * other format having lines containing two positions separated by
 * blanks.  It is also possible to skip lines in the position file
 * that contain specified strings; this makes it easy to avoid
 * masking particular types of repeats, for example, without having
 * to edit the repeats file manually.
 *
 * $Revision: 1.6 $$Date: 2001/04/04 21:39:49 $
 */

#include "seq.h"
#include "feat.h"
#include "util.h"

#define SYNTAX  "seq_file position_file [-char mask_char] " \
                  "[-skip \"string1\" [-skip \"string2\" [...]]] > output_file"

enum { BUFSIZE = 1000, MAXSKIP = 20 };

static void get_args (int argc, char **argv, char **seqfile, char **posfile,
                      char *maskchar, char **skip);

int main (int argc, char **argv) {
	char *seqfile = NULL;
	char *posfile = NULL;
	char maskchar = 0;
	char *skip [MAXSKIP];
	SEQ *sp;                                /* ptr to struct holding seq    */
	uchar *seq;                             /* sequence characters          */
	int seqlen;                             /* length of sequence           */
	FILE *fp;
	char buf [BUFSIZE];
	int width;                              /* number of columns to print   */
	int i;

	/* get arguments */
	set_argv0 (argv[0]);
	for (i = 0; i < MAXSKIP; i++)
		skip[i] = NULL;
	get_args (argc, argv, &seqfile, &posfile, &maskchar, skip);
	if (!seqfile || !posfile)
		fatal (usage (SYNTAX));

	/* read sequence */
	sp = seq_open (seqfile);
	width = seq_get_width (sp);
	if (seq_read (sp) <= 0)
		fatalf ("Couldn't get sequence from \"%s\".", seqfile);
	seq = SEQ_CHARS (sp);
	seqlen = SEQ_LEN (sp);
	if (SEQ_FROM (sp) != 1)
		fatal ("Specification of subsequence is not supported.");

	/* loop over position lines */
	fp = ckopen (posfile, "r");
	while (fgets (buf, BUFSIZE, fp)) {
		char head [BUFSIZE], mid [BUFSIZE], *tail;
		int a1, a2;

		/* get the interval endpoints */
		if (strlen (buf) == BUFSIZE-1)    /* guarantees \n, 1 empty char */
			fatalf ("Line too long in \"%s\":\n%s", posfile, buf);
		for (i = 0; i < MAXSKIP && skip[i]; i++) {
			if (strstr (buf, skip[i]))
				break;
		}
		if (i < MAXSKIP && skip[i])    /* matched a skip string */
			continue;
		if (!feat_parse_line (buf, head, &a1, mid, &a2, &tail))
			continue;
		if (a2 < a1)
			swap (&a1, &a2);
		if (a1 < 1 || a2 > seqlen)
			fatalf ("Mask position is outside sequence range 1..%d:\n%s",
			        seqlen, buf);

		/* do the mask */
		for (i = a1-1; i < a2; i++) {    /* seq starts at 0 */
			seq[i] = (uchar) (maskchar ? maskchar : tolower (seq[i]));
		}
	}

	/* print the results */
	printf (">%s", argv0);
	for (i = 1; i < argc; i++) {
		if (strsame (argv[i-1], "-skip"))
			printf (" \"%s\"", argv[i]);
		else
			printf (" %s", argv[i]);
	}
	printf (": ");   /* no '\n' -- original header will be appended */
	seq_print (sp, width);
	putchar ('\n');

	/* warn if additional (unprocessed) sequences */
	if (seq_read (sp) > 0) {
		fprintf (stderr, "Warning: %s does not support multiple contigs\n"
		         "in the sequence file -- only the first was used.\n", argv0);
	}

	/* clean up */
	fclose (fp);
	seq_close (sp);
	return 0;
}

/* retrieve command-line arguments */
static void get_args (int argc, char **argv, char **seqfile, char **posfile,
                      char *maskchar, char **skip) {
	int i;
	char *p;
	int k = 0;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-char")) {
			if (strlen (p = optval (argc, argv, &i, SYNTAX)) != 1)
				fatal ("Invalid mask character.");
			else
				*maskchar = *p;
		}
		else if (strsame (argv[i], "-skip")) {
			if (k < MAXSKIP)
				skip[k++] = optval (argc, argv, &i, SYNTAX);
			else
				fatalf ("Only %d skip strings allowed.", MAXSKIP);
		}
		else if (!*seqfile)
			*seqfile = argv[i];
		else if (!*posfile)
			*posfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

