/* shift-pos.c -- shift a file's position coords by a given offset,
 *                or convert them to the reverse sequence
 *
 * This program reads a file containing position intervals, and
 * adjusts the positions in one of two ways.  It can either add a
 * given fixed offset to each position, or compute the corresponding
 * position in the reverse complement of the sequence.  Several
 * file formats are supported, including those for exons, repeats,
 * annotations, and underlays, or any other format having lines
 * containing two positions separated by blanks.
 *
 * Lines beginning with '#' and those that do not contain a position
 * interval are left unchanged.  Lines that cannot be converted (e.g.,
 * because one or both of the positions would end up being negative)
 * are flagged with "### FAILED ###" in the output, and are also
 * displayed on the screen (stderr) as a warning.
 *
 * When computing reverse complement positions, the program will
 * also swap the interval endpoints and make other adjustments as
 * follows.  For exons and repeats files, it will interchange the
 * direction indicators ">" vs. "<" and "Right" vs. "Left",
 * respectively.  For exons files, it will also reverse the order
 * of the entries, to keep them in ascending numerical order.  By
 * default the program will attempt to recognize the file format,
 * but this can be overridden by specifying the file type on the
 * command line.
 *
 * $Revision: 1.21 $$Date: 2001/04/19 22:01:36 $
 */

#include "feat.h"
#include "util.h"

#define SYNTAX  "position_file -add offset | -reverse seq_length " \
                  "[-exons | -repeats | -other] > output_file"
#define FAIL    "### FAILED ### "

enum { BUFSIZE = 1000 };

static void get_args (int argc, char **argv, char **posfile, int *offset,
                      int *seqlen, char *filetype);
static int convert (int pos, int offset, int seqlen);
static int skip_headers (char **lines, int start, int end, char filetype);
static char *skip_fail (char *p);
static int find_last_block (char **lines, int start, int end, char filetype);
static int prev_nonempty_line (char **lines, int start, int i);
static void print_range_fwd (char **lines, int start, int end);
static void print_range_rev (char **lines, int start, int end);
static void reverse_exons (char **lines, int start, int end);
static void free_lines (char **lines, int size);

int main (int argc, char **argv) {
	char *posfile = NULL;
	int offset = 0;
	int seqlen = 0;
	char filetype = 0;   /* 'e'xons or 'r'epeats (others don't matter) */
	FILE *fp;
	char buf [BUFSIZE];
	char **lines;        /* array of converted lines */
	int nlines = 0;      /* number of lines (no end marker needed) */
	int lastline;        /* index of last line (for readability) */
	int i;

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &posfile, &offset, &seqlen, &filetype);
	if (!posfile)
		fatal (usage (SYNTAX));
	if ((offset && seqlen) || (!offset && !seqlen))
		fatal ("You must specify either -add or -reverse, but not both.");

	/* allocate array of lines */
	fp = ckopen (posfile, "r");
	while (fgets (buf, BUFSIZE, fp))
		nlines++;
	lines = (char **) ckalloc (nlines * sizeof (char *));
	rewind (fp);

	/* read lines into array; convert positions */
	i = -1;
	while (fgets (buf, BUFSIZE, fp)) {
		char head [BUFSIZE], mid [BUFSIZE], *tail;
		int a1, a2;

		if (strlen (buf) == BUFSIZE-1)    /* guarantees \n, 1 empty char */
			fatalf ("Line too long in \"%s\":\n%s", posfile, buf);
		if (seqlen && !filetype)
			feat_guess_filetype (buf, &filetype);

		if (!feat_parse_line (buf, head, &a1, mid, &a2, &tail)) {
			lines[++i] = (char *) ckalloc (strlen (buf) + 1);
			strcpy (lines[i], buf);
			continue;
		}

		a1 = convert (a1, offset, seqlen);
		a2 = convert (a2, offset, seqlen);

		if (!a1 || !a2) {
			lines[++i] = (char *) ckalloc (strlen (FAIL) + strlen (buf) + 1);
			sprintf (lines[i], "%s%s", FAIL, buf);
			fprintf (stderr, "%s%s", FAIL, buf);
		}
		else {
			if (seqlen) {
				swap (&a1, &a2);
				feat_reverse_dir (head, tail, filetype);
				                                /* buf does have room */
			}
			lines[++i] = (char *) ckalloc (strlen (buf) + 18 + 1);
			sprintf (lines[i], "%s%d%s%d%s", head, a1, mid, a2, tail);
		}
	}
	if (i != nlines-1)
		fatalf ("Internal error: bad line count.");
	fclose (fp);

	/* print lines from array to stdout */
	lastline = nlines-1;
	if (offset || (filetype != 'e')) {
		print_range_fwd (lines, 0, lastline);
	}
	else {    /* reverse */
		int start, end;

		/* header comments */
		start = skip_headers (lines, 0, lastline, filetype);
		print_range_fwd (lines, 0, start-1);

		/* middle lines, with appropriate reversal */
		end = prev_nonempty_line (lines, start, lastline+1);
		if (end >= start) {
			switch (filetype) {
				case 'e': reverse_exons (lines, start, end); break;
				default:  fatalf ("Invalid filetype '%c' for reversal.",
				                  filetype);
			  /*	default:  print_range_rev (lines, start, end); */
			}
		}
		/* trailing blank lines */
		print_range_fwd (lines, end+1, lastline);
	}

	/* clean up */
	free_lines (lines, nlines);
	return 0;
}

/* retrieve command-line arguments */
static void get_args (int argc, char **argv, char **posfile, int *offset,
                      int *seqlen, char *filetype) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-add")) {
			if (!(*offset = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid offset value.");
		}
		else if (strsame (argv[i], "-reverse")) {
			if (!(*seqlen = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid seq_length value.");
		}
		else if (strsame (argv[i], "-exons"))
			*filetype = 'e';
		else if (strsame (argv[i], "-repeats"))
			*filetype = 'r';
		else if (strsame (argv[i], "-other"))
			*filetype = 'o';
		else if (!*posfile)
			*posfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

/* convert a position */
static int convert (int pos, int offset, int seqlen) {
	int n = (offset ? pos + offset : seqlen - pos + 1);
	return (n < 0 ? 0 : n);
}

/* find top of margin preceding the first line of real data */
static int skip_headers (char **lines, int start, int end, char filetype) {
	int i;
	char *p;

	for (i = start; i <= end; i++) {
		p = skip_ws (skip_fail (lines[i]));
		if (filetype == 'e') {
			if (*p == '>' || *p == '<' || *p == '|')
				break;
		}
	}
	for (i--; i >= start; i--) {
		p = skip_ws (skip_fail (lines[i]));
		if (*p)
			break;
	}
	return ++i;    /* range is start .. end+1 */
}

/* disregard FAIL indicator, if present */
static char *skip_fail (char *p) {
	return (starts (p, FAIL) ? p + strlen (FAIL) : p);
}

/* find the first line of the last gene */
static int find_last_block (char **lines, int start, int end, char filetype) {
	int i;
	char *p;

	for (i = end; i >= start; i--) {
		p = skip_ws (skip_fail (lines[i]));
		if (filetype == 'e') {
			if (*p == '>' || *p == '<' || *p == '|')
				break;
		}
		else
			fatalf ("Invalid filetype '%c' in find_last_block().", filetype);
	}
	if (i < start)
		fatalf ("Beginning line of gene is missing.");

	return i;
}

/* find the non-space, non-comment line preceding line i */
static int prev_nonempty_line (char **lines, int start, int i) {
	char *p;

	for (i--; i >= start; i--) {
		p = skip_ws (skip_fail (lines[i]));
		if (*p && *p != '#')
			break;
	}
	return i;    /* not found => start-1 */
}

/* print a range of lines, inclusive */
static void print_range_fwd (char **lines, int start, int end) {
	int i;

	for (i = start; i <= end; i++) {
		fputs (lines[i], stdout);
	}
}

/* print a range of lines, inclusive, in reversed order */
static void print_range_rev (char **lines, int start, int end) {
	int i;

	for (i = end; i >= start; i--) {
		fputs (lines[i], stdout);
	}
}

/* print lines from exons file in reversed order */
static void reverse_exons (char **lines, int start, int end) {
	int g, i;
	char *p;

	while (end >= start) {
		g = find_last_block (lines, start, end, 'e');
		for (i = g+1; i <= end; i++) {
			p = skip_ws (skip_fail (lines[i]));
			if (*p && *p != '+')
				break;
		}
		g = prev_nonempty_line (lines, start, g) + 1;
		print_range_fwd (lines, g, i-1);
		print_range_rev (lines, i, end);
		end = g-1;
	}
}

/* free allocated line array */
static void free_lines (char **lines, int size) {
	int i;

	for (i = 0; i < size; i++)
		free (lines[i]);
	free (lines);
}

